#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

typedef struct quick_sort_args_t {
  int *a;
  int start;
  int end;
} quick_sort_args_t;

static void* do_quick_sort(void *args);

#define do_parallel 1

void quick_sort(int *a, int start, int end)
{
  if (end - start > 1) {
    int pivot = a[start]; /* purple value */
    int pre = start; /* red pointer */
    for (int i /* blue pointer */ = start + 1; i < end; i++) {
      int v = a[i];
      if (v < pivot) {
        a[pre] = v;
        pre++;
        a[i] = a[pre];
      }
    }
    a[pre] = pivot;
    if (do_parallel && (end - start > 10000)) {
      quick_sort_args_t a1, a2;
      pthread_t t1, t2;
      a1.a = a;
      a1.start = start;
      a1.end = pre;
      a2.a = a;
      a2.start = pre+1;
      a2.end = end;
      pthread_create(&t1, NULL, do_quick_sort, &a1);
      pthread_create(&t2, NULL, do_quick_sort, &a2);
      pthread_join(t1, NULL);
      pthread_join(t2, NULL);
    } else {
      quick_sort(a, start, pre);
      quick_sort(a, pre+1, end);
    }
  }
}

static void* do_quick_sort(void *_args)
{
  quick_sort_args_t *args = (quick_sort_args_t *)_args;
  quick_sort(args->a, args->start, args->end);
  return NULL;
}

/*************************************************************/

void check_array(int* a, int* b, int len)
{
  int i;
  for (i = 0; i < len; i++) {
    if (a[i] != b[i]) {
      printf("failed, %d != %d at %d\n", a[i], b[i], i);
      exit(1);
    }
  }
}

void tests() {
  int a1[10] = { 8, 9, 3, 6, 7, 2, 4, 5, 1, 0 };
  int a2[10] = { 9, 8, 7, 6, 5, 4, 3, 2, 1, 0 };
  int a3[10] = { 0, 2, 3, 7, 5, 6, 4, 1, 9, 8 };
  int a4[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
  int s[10] =  { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
  
  quick_sort(a1, 0, 10);
  check_array(a1, s, 10);
  quick_sort(a2, 0, 10);
  check_array(a2, s, 10);
  quick_sort(a3, 0, 10);
  check_array(a3, s, 10);
  quick_sort(a4, 0, 10);
  check_array(a4, s, 10);
}

/*************************************************************/

struct container {
  int count, size;
  int *a;
};

typedef struct container * array_container;

array_container make_array_container() {
  array_container ac;

  ac = (array_container)malloc(sizeof(struct container));
  ac->count = 0;
  ac->size = 1;
  ac->a = (int*)malloc(ac->size * sizeof(int));

  return ac;
}

void copy_array(int* to, int* from, int len) {
  /* copies an array starting from the last item */
  int i;
  for (i = len; i-- > 0; ) {
    to[i] = from[i];
  }
}

void maybe_grow_array(array_container ac) {
  int *new_a;
  if (ac->count >= ac->size) {
    ac->size *= 2;
    new_a = (int*)malloc(ac->size * sizeof(int));
    copy_array(new_a, ac->a, ac->count);
    ac->a = new_a;
  }
}

void add_to_back(array_container ac, int i) {
  maybe_grow_array(ac);

  ac->a[ac->count] = i;
  ac->count++;
}

/*************************************************************/

int main()
{
  array_container ac;
  char buffer[256];
  int i;

  tests();

  ac = make_array_container();
  
  for (; fgets(buffer, 256, stdin); ) {
    add_to_back(ac, atoi(buffer));
  }

  quick_sort(ac->a, 0, ac->count);

  return 1;
}
